﻿#undef debugTestUtilities
#define debugTestUtilities

using CySmart.Common.Base.Compatibility.Net20;
using CySmart.DongleCommunicator.API;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO.Ports;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;


namespace Host
{
    public static class Utilities
    {
        public static byte[] HexStringToByteArray(string rawHex, int requiredByteLen)
        {
            int hiNibble;
            int loNibble;
            string hex;
            bool bigEndian;

            // If the input string contains a space character, then it represents discrete bytes in little-endian format.
            // Remove all space characters so we can convert actual hex characters.
            if (rawHex.IndexOf(" ") != -1)
            {
                rawHex = rawHex.Replace(" ", "");
                bigEndian = false;
            }
            else
            {
                bigEndian = true;
            }

            // Compensate for missing leading '0', so this method won't throw an exception
            if (rawHex.Length % 2 == 1)
            {
                hex = rawHex.PadLeft(rawHex.Length + 1, '0');
            }
            else
            {
                hex = rawHex;
            }

            // If no 'expected size' has been specified, simply convert the entire string
            if (requiredByteLen == 0)
            {
                requiredByteLen = hex.Length / 2;
            }
            byte[] bytes = new byte[requiredByteLen];
            int[] hexValue = new int[]
            {
                0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09,
                0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F
            };

            int b = 0;
            for (int i = 0; (i < hex.Length) && (b < requiredByteLen); i += 2, b++)
            {
                hiNibble = Char.ToUpper(hex[i + 0]) - '0';
                if (hiNibble >= hexValue.Length)
                {
                    return null;
                }
                loNibble = Char.ToUpper(hex[i + 1]) - '0';
                if (loNibble >= hexValue.Length)
                {
                    return null;
                }
                bytes[b] = (byte)(hexValue[hiNibble] << 4 | hexValue[loNibble]);
            }
            // If it's a big-endian input and we have more than 1 byte, reverse the byte order
            if (bigEndian && (b > 1))
            {
                Array.Reverse(bytes, 0, b);
            }
            // Pad right with zero bytes
            for (;  b < requiredByteLen; b++)
            {
                bytes[b] = 0x00;
            }

            return bytes;
        }

        public static string ByteArrayToHexString(byte[] bytes)
        {
            char[] c = new char[3 * bytes.Length - 1];
            byte b;

            for (int i = 0; i < bytes.Length; i++)
            {
                b = (byte)(bytes[i] >> 4);
                c[i * 3] = (char)(b > 9 ? b + 0x37 : b + 0x30);
                b = (byte)(bytes[i] & 0x0F);
                c[i * 3 + 1] = (char)(b > 9 ? b + 0x37 : b + 0x30);
                if (i < bytes.Length - 1)
                {
                    c[i * 3 + 2] = ' ';
                }
            }
            return new string(c);
        }

        // Horribly inefficient, but much more concise than comparing state vs. newState and updating only
        //  those parameters that have actually changed through the state transition
        public static void SetState(Constants.hostStates newState)
        {
            Globals.state = newState;
            switch (Globals.state)
            {
                case Constants.hostStates.HOST_STATE_IDLE:
                    Program.mf.ConfigureButtons(true, "Connect Dongle", false, "Start Scan",
                                                false, "Connect Device", false, false);
                    break;

                case Constants.hostStates.HOST_STATE_CONNECTING_DONGLE:
                    Program.mf.ConfigureButtons(false, "Connect Dongle", false, "Start Scan",
                                                false, "Connect Device", false, false);
                    break;

                case Constants.hostStates.HOST_STATE_READY:
                    Program.mf.ConfigureButtons(true, "Disconnect Dongle", true, "Start Scan",
                                                IsConnectable(Program.mf.scanResults), "Connect Device", false, false);
                    break;

                case Constants.hostStates.HOST_STATE_SCAN_START:
                case Constants.hostStates.HOST_STATE_CONNECTION_PENDING:
                    Program.mf.ConfigureButtons(false, "Disconnect Dongle", false, "Start Scan",
                                                false, "Connect Device", false, false);
                    break;

                case Constants.hostStates.HOST_STATE_SCANNING:
                    Program.mf.ConfigureButtons(false, "Disconnect Dongle", true, "Stop Scan",
                                                IsConnectable(Program.mf.scanResults), "Connect Device", false, false);
                    break;

                case Constants.hostStates.HOST_STATE_SCAN_STOP:
                case Constants.hostStates.HOST_STATE_SCAN_STOP_THEN_CONNECT:
                    Program.mf.ConfigureButtons(false, "Disconnect Dongle", false, "Stop Scan",
                                                false, "Connect Device", false, false);
                    break;

                case Constants.hostStates.HOST_STATE_CONNECTING_DEVICE:
                case Constants.hostStates.HOST_STATE_PAIRING:
                    Program.mf.ConfigureButtons(false, "Disconnect Dongle", false, "Start Scan",
                                                true, "Cancel Connection", false, false);
                    break;

                case Constants.hostStates.HOST_STATE_CANCELING_CONNECTION:
                    Program.mf.ConfigureButtons(false, "Disconnect Dongle", false, "Start Scan",
                                                false, "Cancel Connection", false, false);
                    break;

                case Constants.hostStates.HOST_STATE_DISCOVERING_GATT:
                case Constants.hostStates.HOST_STATE_DISCONNECTING:
                case Constants.hostStates.HOST_STATE_GATT_SHOWN:
                case Constants.hostStates.HOST_STATE_UAPI_SHOWN:
                    Program.mf.ConfigureButtons(false, "Disconnect Dongle", false, "Start Scan",
                                                false, "Disconnect Device", false, false);
                    break;

                case Constants.hostStates.HOST_STATE_DEVICE_CONNECTED:
                    Program.mf.ConfigureButtons(false, "Disconnect Dongle", false, "Start Scan",
                                                true, "Disconnect Device", true, true);
                    break;

                default:
#if (debugTestUtilities)
                    // The following is debug-only nonsense
                    ForceException(13);
#endif
                    break;
            }
        }

        public delegate void ControlStringConsumer(Control control, string text);
        public static void SetText(Control control, string text)
        {
            if (control.InvokeRequired)
            {
                control.Invoke(new ControlStringConsumer(SetText), new object[] { control, text });
            }
            else
            {
                control.Text = text;
            }
        }

        public delegate void ScannedDevicesAdder(ListView listView, CyScanRecord device);
        public static void AddDevice(ListView listView, CyScanRecord device)
        {
            if (listView.InvokeRequired)
            {
                listView.Invoke(new ScannedDevicesAdder(AddDevice), new object[] { listView, device });
            }
            else
            {
                foreach (ListViewItem dev in listView.Items)
                {
                    if ( (((CyScanRecord)dev.Tag).PeerDeviceAddress.Address == device.PeerDeviceAddress.Address) &&
                         (((CyScanRecord)dev.Tag).AdvertisementType == device.AdvertisementType) )
                    {
                        return;
                    }
                }
                ListViewItem listItem = new ListViewItem(string.Format("{0:X12}", device.PeerDeviceAddress.Address));
                listItem.SubItems.Add(device.PeerDeviceAddress.AddressType.ToString());
                listItem.SubItems.Add(device.AdvertisementType.ToString());
                listItem.SubItems.Add(device.RSSI.ToString());
                listItem.Tag = device;
                listView.Items.Add(listItem);
            }
        }

        public delegate bool ScannedDevicesReader(ListView listView);
        public static bool IsConnectable(ListView listView)
        {
            if (listView.InvokeRequired)
            {
                return (bool)listView.Invoke(new ScannedDevicesReader(IsConnectable), new object[] { listView });
            }
            else
            {
                if (listView.SelectedItems.Count > 0)
                {
                    Program.mf.highlightedDevice = (CyScanRecord)listView.SelectedItems[0].Tag;
                    if ((Program.mf.highlightedDevice.AdvertisementType == CySmart.Common.Base.CyBleAdvEventType.CONNECTABLE_DIRECTED_ADVERTISING) ||
                         (Program.mf.highlightedDevice.AdvertisementType == CySmart.Common.Base.CyBleAdvEventType.CONNECTABLE_UNDIRECTED_ADVERTISING))
                    {
                        return true;
                    }
                }
                return false;
            }
        }

        public delegate void ListViewClearer(ListView listView);
        public static void ClearListView(ListView listView)
        {
            if (listView.InvokeRequired)
            {
                listView.Invoke(new ListViewClearer(ClearListView), new object[] { listView });
            }
            else
            {
                if (listView.Items.Count > 0)
                {
                    listView.Items.Clear();
                }
            }
        }

        public delegate void ButtonEnabler(Button button, bool enabled);
        public static void EnableButton(Button button, bool enabled)
        {
            if (button.InvokeRequired)
            {
                button.Invoke(new ButtonEnabler(EnableButton), new object[] { button, enabled });
            }
            else
            {
                button.Enabled = enabled;
            }
        }

#if(debugTestUtilities)
        public static void ForceException(int which)
        {
            try
            {
                throw new DivideByZeroException();
            }
            catch (DivideByZeroException ex)
            {
                string sm = string.Format("Exception{0}: {1}", which, ex);
                SetText(Program.mf.statusMessage, sm);
            }
        }
#endif

        public static byte[] CreateSpecialByteArray(int length, byte val)
        {
            var arr = new byte[length];

            for (int i = 0; i < arr.Length; i++)
            {
                arr[i] = val;
            }
            return arr;
        }
    }

    public static class RichTextBoxExtensions
    {
        public delegate void TextBoxAppender(RichTextBox box, string text, Color color);
        public static void AppendText(this RichTextBox box, string text, Color color)
        {
            if (box.InvokeRequired)
            {
                box.Invoke(new TextBoxAppender(AppendText), new object[] { box, text, color });
            }
            else
            {
                box.SelectionStart = box.TextLength;
                box.SelectionLength = 0;
                box.SelectionColor = color;
                box.AppendText(text);
                box.SelectionColor = box.ForeColor;
            }
        }
    }

#if (useLinkedLists)
    public class Node
    {
        public Node next;
        public Object data;
    }

    public class LinkedList
    {
        private Node head = null;

        public void Add(Object data)
        {
            if (head == null)
            {
                head = new Node();

                head.data = data;
                head.next = null;
            }
            else
            {
                Node toAdd = new Node();
                toAdd.data = data;
                toAdd.next = null;

                Node current = head;
                while (current.next != null)
                {
                    current = current.next;
                }

                current.next = toAdd;
            }
        }

        public Node GetNode(int index)
        {
            Node node = head;

            for (int count = 0; (count < index) && (node != null); count++)
            {
                node = node.next;
            }
            return node;
        }
    }
#endif
}
